@@ -3,71 +3,64 @@
 # Copyright (c) Jupyter Development Team
 # Distributed under the terms of the Modified BSD License.
 
-from __future__ import print_function
-
 from collections import namedtuple
 import sys
 import uuid
-try:
-    from queue import Empty
-except ImportError:
-    from Queue import Empty
 import re
 
-from qtconsole import qt
-from qtconsole.qt import QtCore, QtGui
-from ipython_genutils import py3compat
+from qtpy import QtCore, QtGui, QtWidgets
 from ipython_genutils.importstring import import_item
 
 from qtconsole.base_frontend_mixin import BaseFrontendMixin
-from traitlets import Any, Bool, Float, Instance, Unicode, DottedObjectName
+from traitlets import Any, Bool, Instance, Unicode, DottedObjectName, default
 from .bracket_matcher import BracketMatcher
 from .call_tip_widget import CallTipWidget
 from .history_console_widget import HistoryConsoleWidget
 from .pygments_highlighter import PygmentsHighlighter
 
-_classic_prompt_re = re.compile(r'^([ \t]*>>> |^[ \t]*\.\.\. )')
-
-def transform_classic_prompt(line):
-    """Handle inputs that start with '>>> ' syntax."""
-
-    if not line or line.isspace():
-        return line
-    m = _classic_prompt_re.match(line)
-    if m:
-        return line[len(m.group(0)):]
-    else:
-        return line
-
-
-_ipy_prompt_re = re.compile(r'^([ \t]*In \[\d+\]: |^[ \t]*\ \ \ \.\.\.+: )')
-
-def transform_ipy_prompt(line):
-    """Handle inputs that start classic IPython prompt syntax."""
-
-    if not line or line.isspace():
-        return line
-    m = _ipy_prompt_re.match(line)
-    if m:
-        return line[len(m.group(0)):]
-    else:
-        return line
-
 
 class FrontendHighlighter(PygmentsHighlighter):
     """ A PygmentsHighlighter that understands and ignores prompts.
     """
 
     def __init__(self, frontend, lexer=None):
-        super(FrontendHighlighter, self).__init__(frontend._control.document(), lexer=lexer)
+        super().__init__(frontend._control.document(), lexer=lexer)
         self._current_offset = 0
         self._frontend = frontend
         self.highlighting_on = False
+        self._classic_prompt_re = re.compile(
+            r'^(%s)?([ \t]*>>> |^[ \t]*\.\.\. )' % re.escape(frontend.other_output_prefix)
+        )
+        self._ipy_prompt_re = re.compile(
+            r'^(%s)?([ \t]*In \[\d+\]: |[ \t]*\ \ \ \.\.\.+: )' % re.escape(frontend.other_output_prefix)
+        )
+
+    def transform_classic_prompt(self, line):
+        """Handle inputs that start with '>>> ' syntax."""
+
+        if not line or line.isspace():
+            return line
+        m = self._classic_prompt_re.match(line)
+        if m:
+            return line[len(m.group(0)):]
+        else:
+            return line
+
+    def transform_ipy_prompt(self, line):
+        """Handle inputs that start classic IPython prompt syntax."""
+
+        if not line or line.isspace():
+            return line
+        m = self._ipy_prompt_re.match(line)
+        if m:
+            return line[len(m.group(0)):]
+        else:
+            return line
 
     def highlightBlock(self, string):
         """ Highlight a block of text. Reimplemented to highlight selectively.
         """
-        if not self.highlighting_on:
+        if not hasattr(self, 'highlighting_on') or not self.highlighting_on:
             return
 
         # The input to this function is a unicode string that may contain
@@ -76,32 +69,27 @@ class FrontendHighlighter(PygmentsHighlighter):
         current_block = self.currentBlock()
         string = self._frontend._get_block_plain_text(current_block)
 
-        # Decide whether to check for the regular or continuation prompt.
-        if current_block.contains(self._frontend._prompt_pos):
-            prompt = self._frontend._prompt
-        else:
-            prompt = self._frontend._continuation_prompt
-
         # Only highlight if we can identify a prompt, but make sure not to
         # highlight the prompt.
-        if string.startswith(prompt):
-            self._current_offset = len(prompt)
-            string = string[len(prompt):]
-            super(FrontendHighlighter, self).highlightBlock(string)
+        without_prompt = self.transform_ipy_prompt(string)
+        diff = len(string) - len(without_prompt)
+        if diff > 0:
+            self._current_offset = diff
+            super().highlightBlock(without_prompt)
 
     def rehighlightBlock(self, block):
         """ Reimplemented to temporarily enable highlighting if disabled.
         """
         old = self.highlighting_on
         self.highlighting_on = True
-        super(FrontendHighlighter, self).rehighlightBlock(block)
+        super().rehighlightBlock(block)
         self.highlighting_on = old
 
     def setFormat(self, start, count, format):
         """ Reimplemented to highlight selectively.
         """
         start += self._current_offset
-        super(FrontendHighlighter, self).setFormat(start, count, format)
+        super().setFormat(start, count, format)
 
 
 class FrontendWidget(HistoryConsoleWidget, BaseFrontendMixin):
@@ -134,23 +122,17 @@ class FrontendWidget(HistoryConsoleWidget, BaseFrontendMixin):
 
     confirm_restart = Bool(True, config=True,
         help="Whether to ask for user confirmation when restarting kernel")
-    
+
     lexer_class = DottedObjectName(config=True,
         help="The pygments lexer class to use."
     )
-    is_complete_timeout = Float(0.25, config=True,
-        help="Seconds to wait for is_complete replies from the kernel."
-    )
     def _lexer_class_changed(self, name, old, new):
         lexer_class = import_item(new)
         self.lexer = lexer_class()
-    
+
     def _lexer_class_default(self):
-        if py3compat.PY3:
-            return 'pygments.lexers.Python3Lexer'
-        else:
-            return 'pygments.lexers.PythonLexer'
-    
+        return 'pygments.lexers.Python3Lexer'
+
     lexer = Any()
     def _lexer_default(self):
         lexer_class = import_item(self.lexer_class)
@@ -168,29 +150,22 @@ class FrontendWidget(HistoryConsoleWidget, BaseFrontendMixin):
     exit_requested = QtCore.Signal(object)
 
     _CallTipRequest = namedtuple('_CallTipRequest', ['id', 'pos'])
-    _CompletionRequest = namedtuple('_CompletionRequest', ['id', 'pos'])
+    _CompletionRequest = namedtuple('_CompletionRequest',
+                                    ['id', 'code', 'pos'])
     _ExecutionRequest = namedtuple('_ExecutionRequest', ['id', 'kind'])
     _local_kernel = False
     _highlighter = Instance(FrontendHighlighter, allow_none=True)
 
-    #---------------------------------------------------------------------------
-    # 'object' interface
-    #---------------------------------------------------------------------------
-
-    def __init__(self, *args, **kw):
-        super(FrontendWidget, self).__init__(*args, **kw)
-        # FIXME: remove this when PySide min version is updated past 1.0.7
-        # forcefully disable calltips if PySide is < 1.0.7, because they crash
-        if qt.QT_API == qt.QT_API_PYSIDE:
-            import PySide
-            if PySide.__version_info__ < (1,0,7):
-                self.log.warn("PySide %s < 1.0.7 detected, disabling calltips" % PySide.__version__)
-                self.enable_calltips = False
+    # -------------------------------------------------------------------------
+    # 'Object' interface
+    # -------------------------------------------------------------------------
 
+    def __init__(self, local_kernel=_local_kernel, *args, **kw):
+        super().__init__(*args, **kw)
         # FrontendWidget protected variables.
         self._bracket_matcher = BracketMatcher(self._control)
         self._call_tip_widget = CallTipWidget(self._control)
-        self._copy_raw_action = QtGui.QAction('Copy (Raw Text)', None)
+        self._copy_raw_action = QtWidgets.QAction('Copy (Raw Text)', None)
         self._hidden = False
         self._highlighter = FrontendHighlighter(self, lexer=self.lexer)
         self._kernel_manager = None
@@ -223,8 +198,7 @@ class FrontendWidget(HistoryConsoleWidget, BaseFrontendMixin):
         document.contentsChange.connect(self._document_contents_change)
 
         # Set flag for whether we are connected via localhost.
-        self._local_kernel = kw.get('local_kernel',
-                                    FrontendWidget._local_kernel)
+        self._local_kernel = local_kernel
 
         # Whether or not a clear_output call is pending new output.
         self._pending_clearoutput = False
@@ -241,15 +215,61 @@ class FrontendWidget(HistoryConsoleWidget, BaseFrontendMixin):
         elif self._control.hasFocus():
             text = self._control.textCursor().selection().toPlainText()
             if text:
-                # Remove prompts.
-                lines = text.splitlines()
-                lines = map(transform_classic_prompt, lines)
-                lines = map(transform_ipy_prompt, lines)
-                text = '\n'.join(lines)
-                was_newline = text[-1] == '\n'
+                first_line_selection, *remaining_lines = text.splitlines()
+
+                # Get preceding text
+                cursor = self._control.textCursor()
+                cursor.setPosition(cursor.selectionStart())
+                cursor.setPosition(cursor.block().position(),
+                                   QtGui.QTextCursor.KeepAnchor)
+                preceding_text = cursor.selection().toPlainText()
+
+                def remove_prompts(line):
+                    """Remove all prompts from line."""
+                    line = self._highlighter.transform_classic_prompt(line)
+                    return self._highlighter.transform_ipy_prompt(line)
+
+                # Get first line promp len
+                first_line = preceding_text + first_line_selection
+                len_with_prompt = len(first_line)
+                first_line = remove_prompts(first_line)
+                prompt_len = len_with_prompt - len(first_line)
+
+                # Remove not selected part
+                if prompt_len < len(preceding_text):
+                    first_line = first_line[len(preceding_text) - prompt_len:]
+
+                # Remove partial prompt last line
+                if len(remaining_lines) > 0 and remaining_lines[-1]:
+                    cursor = self._control.textCursor()
+                    cursor.setPosition(cursor.selectionEnd())
+                    block = cursor.block()
+                    start_pos = block.position()
+                    length = block.length()
+                    cursor.setPosition(start_pos)
+                    cursor.setPosition(start_pos + length - 1,
+                                       QtGui.QTextCursor.KeepAnchor)
+                    last_line_full = cursor.selection().toPlainText()
+                    prompt_len = (
+                        len(last_line_full)
+                        - len(remove_prompts(last_line_full)))
+                    if len(remaining_lines[-1]) < prompt_len:
+                        # This is a partial prompt
+                        remaining_lines[-1] = ""
+
+                # Remove prompts for other lines.
+                remaining_lines = map(remove_prompts, remaining_lines)
+                text = '\n'.join([first_line, *remaining_lines])
+
+                # Needed to prevent errors when copying the prompt.
+                # See issue 264
+                try:
+                    was_newline = text[-1] == '\n'
+                except IndexError:
+                    was_newline = False
                 if was_newline:  # user doesn't need newline
                     text = text[:-1]
-                QtGui.QApplication.clipboard().setText(text)
+                QtWidgets.QApplication.clipboard().setText(text)
         else:
             self.log.debug("frontend widget : unknown copy target")
 
@@ -257,35 +277,6 @@ class FrontendWidget(HistoryConsoleWidget, BaseFrontendMixin):
     # 'ConsoleWidget' abstract interface
     #---------------------------------------------------------------------------
 
-    def _is_complete(self, source, interactive):
-        """ Returns whether 'source' can be completely processed and a new
-            prompt created. When triggered by an Enter/Return key press,
-            'interactive' is True; otherwise, it is False.
-            
-            Returns
-            -------
-            
-            (complete, indent): (bool, str)
-            complete is a bool, indicating whether the input is complete or not.
-            indent is the current indentation string for autoindent.
-            If complete is True, indent will be '', and should be ignored.
-        """
-        kc = self.blocking_client
-        if kc is None:
-            self.log.warn("No blocking client to make is_complete requests")
-            return False, u''
-        msg_id = kc.is_complete(source)
-        while True:
-            try:
-                reply = kc.shell_channel.get_msg(block=True, timeout=self.is_complete_timeout)
-            except Empty:
-                # assume incomplete output if we get no reply in time
-                return False, u''
-            if reply['parent_header'].get('msg_id', None) == msg_id:
-                status = reply['content'].get('status', u'complete')
-                indent = reply['content'].get('indent', u'')
-                return status != 'incomplete', indent
-
     def _execute(self, source, hidden):
         """ Execute 'source'. If 'hidden', do not show any output.
 
@@ -317,10 +308,12 @@ class FrontendWidget(HistoryConsoleWidget, BaseFrontendMixin):
         # Perform tab completion if:
         # 1) The cursor is in the input buffer.
         # 2) There is a non-whitespace character before the cursor.
+        # 3) There is no active selection.
         text = self._get_input_buffer_cursor_line()
         if text is None:
             return False
-        complete = bool(text[:self._get_input_buffer_cursor_column()].strip())
+        non_ws_before = bool(text[:self._get_input_buffer_cursor_column()].strip())
+        complete = non_ws_before and self._get_cursor().selectedText() == ''
         if complete:
             self._complete()
         return not complete
@@ -332,7 +325,7 @@ class FrontendWidget(HistoryConsoleWidget, BaseFrontendMixin):
     def _context_menu_make(self, pos):
         """ Reimplemented to add an action for raw copy.
         """
-        menu = super(FrontendWidget, self)._context_menu_make(pos)
+        menu = super()._context_menu_make(pos)
         for before_action in menu.actions():
             if before_action.shortcut().matches(QtGui.QKeySequence.Paste) == \
                     QtGui.QKeySequence.ExactMatch:
@@ -355,7 +348,11 @@ class FrontendWidget(HistoryConsoleWidget, BaseFrontendMixin):
         if self._control_key_down(event.modifiers(), include_command=False):
 
             if key == QtCore.Qt.Key_C and self._executing:
-                self.request_interrupt_kernel()
+                # If text is selected, the user probably wants to copy it.
+                if self.can_copy() and event.matches(QtGui.QKeySequence.Copy):
+                    self.copy()
+                else:
+                    self.request_interrupt_kernel()
                 return True
 
             elif key == QtCore.Qt.Key_Period:
@@ -378,14 +375,7 @@ class FrontendWidget(HistoryConsoleWidget, BaseFrontendMixin):
                         cursor.removeSelectedText()
                         return True
 
-        return super(FrontendWidget, self)._event_filter_console_keypress(event)
-
-    def _insert_continuation_prompt(self, cursor, indent=''):
-        """ Reimplemented for auto-indentation.
-        """
-        super(FrontendWidget, self)._insert_continuation_prompt(cursor)
-        if indent:
-            cursor.insertText(indent)
+        return super()._event_filter_console_keypress(event)
 
     #---------------------------------------------------------------------------
     # 'BaseFrontendMixin' abstract interface
@@ -447,9 +437,7 @@ class FrontendWidget(HistoryConsoleWidget, BaseFrontendMixin):
         of the value of corresponding `user_expressions` as argument.
         `callback` is then removed from the known list so that any message
         coming again with the same id won't trigger it.
-
         """
-
         user_exp = msg['content'].get('user_expressions')
         if not user_exp:
             return
@@ -460,7 +448,7 @@ class FrontendWidget(HistoryConsoleWidget, BaseFrontendMixin):
     def _handle_execute_reply(self, msg):
         """ Handles replies for code execution.
         """
-        self.log.debug("execute: %s", msg.get('content', ''))
+        self.log.debug("execute_reply: %s", msg.get('content', ''))
         msg_id = msg['parent_header']['msg_id']
         info = self._request_info['execute'].get(msg_id)
         # unset reading flag, because if execute finished, raw_input can't
@@ -511,6 +499,7 @@ class FrontendWidget(HistoryConsoleWidget, BaseFrontendMixin):
         self.kernel_client.iopub_channel.flush()
 
         def callback(line):
+            self._finalize_input_request()
             self.kernel_client.input(line)
         if self._reading:
             self.log.debug("Got second input request, assuming first was interrupted.")
@@ -526,7 +515,7 @@ class FrontendWidget(HistoryConsoleWidget, BaseFrontendMixin):
     def _handle_kernel_died(self, since_last_heartbeat):
         """Handle the kernel's death (if we do not own the kernel).
         """
-        self.log.warn("kernel died: %s", since_last_heartbeat)
+        self.log.warning("kernel died: %s", since_last_heartbeat)
         if self.custom_restart:
             self.custom_restart_kernel_died.emit(since_last_heartbeat)
         else:
@@ -538,7 +527,7 @@ class FrontendWidget(HistoryConsoleWidget, BaseFrontendMixin):
 
         There's nothing to do but show a message.
         """
-        self.log.warn("kernel restarted")
+        self.log.warning("kernel restarted")
         self._kernel_restarted_message(died=died)
         self.reset()
 
@@ -589,11 +578,11 @@ class FrontendWidget(HistoryConsoleWidget, BaseFrontendMixin):
                     self.exit_requested.emit(self)
                 else:
                     title = self.window().windowTitle()
-                    reply = QtGui.QMessageBox.question(self, title,
+                    reply = QtWidgets.QMessageBox.question(self, title,
                         "Kernel has been shutdown permanently. "
                         "Close the Console?",
-                        QtGui.QMessageBox.Yes,QtGui.QMessageBox.No)
-                    if reply == QtGui.QMessageBox.Yes:
+                        QtWidgets.QMessageBox.Yes,QtWidgets.QMessageBox.No)
+                    if reply == QtWidgets.QMessageBox.Yes:
                         self.exit_requested.emit(self)
 
     def _handle_status(self, msg):
@@ -628,7 +617,7 @@ class FrontendWidget(HistoryConsoleWidget, BaseFrontendMixin):
 
     def interrupt_kernel(self):
         """ Attempts to interrupt the running kernel.
-        
+
         Also unsets _reading flag, to avoid runtime errors
         if raw_input is called again.
         """
@@ -687,10 +676,10 @@ class FrontendWidget(HistoryConsoleWidget, BaseFrontendMixin):
             # they decline. (If they accept, the heartbeat will be un-paused
             # automatically when the kernel is restarted.)
             if self.confirm_restart:
-                buttons = QtGui.QMessageBox.Yes | QtGui.QMessageBox.No
-                result = QtGui.QMessageBox.question(self, 'Restart kernel?',
+                buttons = QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No
+                result = QtWidgets.QMessageBox.question(self, 'Restart kernel?',
                                                     message, buttons)
-                do_restart = result == QtGui.QMessageBox.Yes
+                do_restart = result == QtWidgets.QMessageBox.Yes
             else:
                 # confirm_restart is False, so we don't need to ask user
                 # anything, just do the restart
@@ -718,12 +707,7 @@ class FrontendWidget(HistoryConsoleWidget, BaseFrontendMixin):
 
     def append_stream(self, text):
         """Appends text to the output stream."""
-        # Most consoles treat tabs as being 8 space characters. Convert tabs
-        # to spaces so that output looks as expected regardless of this
-        # widget's tab width.
-        text = text.expandtabs(8)
         self._append_plain_text(text, before_prompt=True)
-        self._control.moveCursor(QtGui.QTextCursor.End)
 
     def flush_clearoutput(self):
         """If a clearoutput is pending, execute it."""
@@ -742,10 +726,10 @@ class FrontendWidget(HistoryConsoleWidget, BaseFrontendMixin):
     #---------------------------------------------------------------------------
     # 'FrontendWidget' protected interface
     #---------------------------------------------------------------------------
-    
+
     def _auto_call_tip(self):
         """Trigger call tip automatically on open parenthesis
-        
+
         Call tips can be requested explcitly with `_call_tip`.
         """
         cursor = self._get_cursor()
@@ -753,7 +737,7 @@ class FrontendWidget(HistoryConsoleWidget, BaseFrontendMixin):
         if cursor.document().characterAt(cursor.position()) == '(':
             # trigger auto call tip on open paren
             self._call_tip()
-    
+
     def _call_tip(self):
         """Shows a call tip, if appropriate, at the current cursor location."""
         # Decide if it makes sense to show a call tip
@@ -770,13 +754,11 @@ class FrontendWidget(HistoryConsoleWidget, BaseFrontendMixin):
     def _complete(self):
         """ Performs completion at the current cursor location.
         """
+        code = self.input_buffer
+        cursor_pos = self._get_input_buffer_cursor_pos()
         # Send the completion request to the kernel
-        msg_id = self.kernel_client.complete(
-            code=self.input_buffer,
-            cursor_pos=self._get_input_buffer_cursor_pos(),
-        )
-        pos = self._get_cursor().position()
-        info = self._CompletionRequest(msg_id, pos)
+        msg_id = self.kernel_client.complete(code=code, cursor_pos=cursor_pos)
+        info = self._CompletionRequest(msg_id, code, cursor_pos)
         self._request_info['complete'] = info
 
     def _process_execute_abort(self, msg):
@@ -841,6 +823,7 @@ class FrontendWidget(HistoryConsoleWidget, BaseFrontendMixin):
 
     #------ Trait default initializers -----------------------------------------
 
+    @default('banner')
     def _banner_default(self):
         """ Returns the standard Python banner.
         """
